/*******************************************************************************
 * Copyright (c) 2009 Pafsanias Ftakas.
 *
 * This file is part of DistProc.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * DistProc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Eclipse Public License v1.0 for more details.
 *
 * Contributors:
 *     Pafsanias Ftakas - initial API and implementation
 *******************************************************************************/
package com.koutra.dist.proc.sink;

import java.io.BufferedOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.concurrent.BlockingDeque;
import java.util.concurrent.Future;
import java.util.concurrent.LinkedBlockingDeque;

import org.apache.log4j.Logger;

import com.koutra.dist.proc.model.ContentType;
import com.koutra.dist.proc.model.IFaucet;
import com.koutra.dist.proc.model.IMuxSink;
import com.koutra.dist.proc.model.IPipelineItem;
import com.koutra.dist.proc.model.ISinkTemplate;
import com.koutra.dist.proc.model.XformationException;
import com.koutra.dist.proc.util.ByteArrayUtil;
import com.koutra.dist.proc.util.ExecutorsHelper;
 
public class StreamMuxSink implements IMuxSink {
	private static final Logger logger = Logger.getLogger(StreamMuxSink.class);
	
	protected enum Type {
		File,
		OutputStream;
	};
	
	protected static class DequePayload {
		public IFaucet faucet;
		public InputStream inputStream;

		@Override
		public boolean equals(Object obj) {
			if (!(obj instanceof DequePayload))
				return false;
			
			DequePayload other = (DequePayload) obj;
			if (inputStream == null && other.inputStream != null)
				return false;
			return faucet.equals(other.faucet) && inputStream.equals(other.inputStream);
		}
	}
	
	protected String id;
	protected IFaucet faucetTemplate;
	protected BlockingDeque<DequePayload> deque;
	protected IFaucet faucet;
	protected boolean hookedUp;
	protected Type type;
	protected OutputStream outputStream;
	protected String path;
	
	/**
	 * @deprecated Use any of the initializing constructors instead.
	 */
	public StreamMuxSink() { }
	
	public StreamMuxSink(String id, OutputStream os) {
		this.id = id;
		this.faucetTemplate = null;
		this.deque = new LinkedBlockingDeque<DequePayload>();
		this.faucet = null;
		this.hookedUp = false;
		this.type = Type.OutputStream;
		this.outputStream = os;
		this.path = null;
	}

	public StreamMuxSink(String id, String path) {
		this.id = id;
		this.faucetTemplate = null;
		this.deque = new LinkedBlockingDeque<DequePayload>();
		this.faucet = null;
		this.hookedUp = false;
		this.type = Type.File;
		this.outputStream = null;
		this.path = path;
	}

	@Override
	public String getId() {
		return id;
	}

	@Override
	public IFaucet getFaucet() {
		DequePayload payload = deque.peek();
		return payload == null ? null : payload.faucet;
	}

	@Override
	public boolean supportsInput(ContentType contentType) {
		switch (contentType) {
		case ByteStream:
			return true;
		case CharStream:
		case XML:
		case ResultSet:
		default:
			return false;
		}
	}

	@Override
	public void hookupFaucet(IFaucet faucet) {
		if (!faucet.supportsOutput(ContentType.ByteStream))
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' must support the ByteStream content type.");
		if (!(faucet instanceof ISinkTemplate))
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' must implement the ISinkTemplate interface.");
		if (!((ISinkTemplate) faucet).isClone()) {
			if (faucetTemplate != null && faucetTemplate == faucet) {
				// Effectively a NO-OP
				return;
			}
			if (faucetTemplate != null)
				throw new XformationException("Hooking a faucet template in an " +
						"already set mux faucet");
			faucetTemplate = faucet;
			faucetTemplate.hookupSink(this);
			return;
		}
		if (hookedUp && faucet == this.faucet) {
			// Effectively a NO-OP.
			return;
		}
		
		this.faucet = faucet;
		this.hookedUp = true;
		
		faucet.hookupSink(this);
	}
	
	@Override
	public void registerSource(Object source) {
		InputStream inputStream = (InputStream) source;
		
		DequePayload payload = new DequePayload();
		payload.faucet = faucet;
		payload.inputStream = inputStream;

		if (deque.contains(payload))
			return;
		
		if (logger.isTraceEnabled())
			logger.trace("Registering input stream: " + inputStream + " with the mux deque");
		
		deque.addLast(payload);
		
		if (inputStream != null) {
			if (faucet instanceof IPipelineItem) {
				((IPipelineItem) faucet).consume(this);
			}
		}
	}
	
	@Override
	public String dumpPipeline() {
		DequePayload payload = deque.peekFirst();
		return getClass().getName() + ": " +
			(payload == null ? "null" : payload.inputStream) + "->" + outputStream;
	}

	@Override
	public void dispose() {
		switch (type) {
		case File:
			try {
				outputStream.close();
			} catch (IOException e) {
				throw new XformationException("Unable to close output stream", e);
			}
			break;
		case OutputStream:
			// Caller owns the output stream. It is their responsibility to close it.
			break;
		}
	}

	@Override
	public void consume() {
		if (!hookedUp && faucetTemplate == null)
			throw new XformationException("Sink has not been set up correctly: " +
					"faucet has not been set");
		
		switch (type) {
		case File:
			try {
				outputStream = new BufferedOutputStream(new FileOutputStream(path));
			} catch (FileNotFoundException e) {
				throw new XformationException("Unable to create output stream", e);
			}
			break;
		case OutputStream:
			break;
		}
		
		if (!hookedUp) {
			// This is an initial run up the chain. Propagate the consumption of the
			// event. This will start filling up the faucet and input stream deque.
			if (faucetTemplate instanceof IPipelineItem) {
				((IPipelineItem) faucetTemplate).consume(this);
			}
		}
			
		try {
			int count;
			byte[] buffer = new byte[8 * 1024];
			while (true) {
				DequePayload payload = null;
				try {
					payload = deque.takeFirst();
				} catch (InterruptedException ie) { }
				if (payload == null)
					break;
				
				IFaucet faucet = payload.faucet;
				InputStream inputStream = payload.inputStream;
				
				if (logger.isTraceEnabled())
					logger.trace("Removed input stream: " + inputStream + " from the deque.");
				
				// If a null has been sent down to the deque then we must exit.
				if (inputStream == null)
					break;
			
				if (logger.isTraceEnabled())
					logger.trace("Using the reader " + inputStream + " in the mux sink");
				while ((count = inputStream.read(buffer)) != -1) {
					if (logger.isTraceEnabled()) {
						String readInput = ByteArrayUtil.getHexDump(buffer, 0, count);
						logger.trace("Read " + readInput + " from input stream " +
								inputStream);
					}
					outputStream.write(buffer, 0, count);
				}
				
				faucet.dispose();
			}
			outputStream.close();
			faucetTemplate.dispose();
		} catch (IOException ioe) {
			logger.error("Error while consuming input", ioe);
			throw new XformationException("Unable to transform stream", ioe);
		}
	}

	@Override
	public Future<?> consumeAsynchronously() {
		return ExecutorsHelper.getInstance().submitInProc(new Runnable() {
			@Override
			public void run() {
				try {
					consume();
				} catch (Throwable t) {
					logger.error("Unable to consume faucet", t);
				}
			}
		});
	}

	@Override
	public void readFrom(DataInputStream in) throws IOException,
			IllegalAccessException, InstantiationException {
		this.id = in.readUTF();
		this.faucetTemplate = null;
		this.deque = new LinkedBlockingDeque<DequePayload>();
		this.faucet = null;
		this.hookedUp = false;
		this.type = Type.File;
		this.outputStream = null;
		this.path = in.readUTF();
	}

	@Override
	public void writeTo(DataOutputStream out) throws IOException {
		if (type != Type.File)
			throw new XformationException(
					"Only StreamMuxSinks of type Type.File can be streamed.");
		
		out.writeUTF(id);
		out.writeUTF(path);
	}

}
