/*******************************************************************************
 * Copyright (c) 2009 Pafsanias Ftakas.
 *
 * This file is part of DistProc.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * DistProc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Eclipse Public License v1.0 for more details.
 *
 * Contributors:
 *     Pafsanias Ftakas - initial API and implementation
 *******************************************************************************/
package com.koutra.dist.proc.sink;

import java.io.BufferedOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.concurrent.Future;

import org.apache.log4j.Logger;

import com.koutra.dist.proc.model.ContentType;
import com.koutra.dist.proc.model.IDemuxFaucet;
import com.koutra.dist.proc.model.IFaucet;
import com.koutra.dist.proc.model.IFaucetTemplate;
import com.koutra.dist.proc.model.IPipelineItem;
import com.koutra.dist.proc.model.ISink;
import com.koutra.dist.proc.model.XformationException;
import com.koutra.dist.proc.util.ByteArrayUtil;
import com.koutra.dist.proc.util.ExecutorsHelper;

/**
 * A sink implementation that writes out to an output stream. When using the
 * constructor that supplies the output stream, the caller has the responsibility
 * of closing the output stream when the transformation is completed.
 * 
 * @author Pafsanias Ftakas
 */
public class StreamSink implements ISink {
	private static final Logger logger = Logger.getLogger(StreamSink.class);
	
	protected enum Type {
		File,
		OutputStream;
	};
	
	protected String id;
	protected IFaucet faucet;
	protected boolean hookedUp;
	protected Type type;
	protected OutputStream outputStream;
	protected InputStream inputStream;
	protected String path;
	
	/**
	 * @deprecated Use any of the initializing constructors instead.
	 */
	public StreamSink() { }
	
	public StreamSink(String id, OutputStream os) {
		this.id = id;
		this.faucet = null;
		this.hookedUp = false;
		this.type = Type.OutputStream;
		this.outputStream = os;
		this.inputStream = null;
		this.path = null;
	}

	public StreamSink(String id, String path) {
		this.id = id;
		this.faucet = null;
		this.hookedUp = false;
		this.type = Type.File;
		this.outputStream = null;
		this.inputStream = null;
		this.path = path;
	}

	@Override
	public String getId() {
		return id;
	}

	@Override
	public IFaucet getFaucet() {
		return faucet;
	}

	@Override
	public boolean supportsInput(ContentType contentType) {
		switch (contentType) {
		case ByteStream:
			return true;
		case CharStream:
		case XML:
		case ResultSet:
		default:
			return false;
		}
	}

	@Override
	public void hookupFaucet(IFaucet faucet) {
		if (!faucet.supportsOutput(ContentType.ByteStream))
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' must support the ByteStream content type.");
		if (faucet instanceof IDemuxFaucet)
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' is an IDemuxFaucet instance. This is not allowed.");
		if (faucet instanceof IFaucetTemplate)
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' is an IFaucetTemplate instance. This is not allowed.");
		if (hookedUp && this.faucet == faucet) {
			// Effectively a NO-OP.
			return;
		}
		if (hookedUp)
			throw new XformationException("Trying to hook up an already hooked up sink.");
		
		this.faucet = faucet;
		this.hookedUp = true;
		
		this.faucet.hookupSink(this);
	}
	
	@Override
	public void registerSource(Object source) {
		// Only used in mux/demux situations.
	}
	
	@Override
	public String dumpPipeline() {
		return getClass().getName() + ": " + inputStream + "->" + outputStream;
	}

	@Override
	public void dispose() {
		switch (type) {
		case File:
			try {
				outputStream.close();
			} catch (IOException e) {
				throw new XformationException("Unable to close output stream", e);
			}
			break;
		case OutputStream:
			// Caller owns the output stream. It is their responsibility to close it.
			break;
		}
		
		faucet.dispose();
	}

	@Override
	public void consume() {
		if (!hookedUp)
			throw new XformationException("Sink has not been set up correctly: " +
					"faucet has not been set");
		
		switch (type) {
		case File:
			try {
				outputStream = new BufferedOutputStream(new FileOutputStream(path));
			} catch (FileNotFoundException e) {
				throw new XformationException("Unable to create output stream", e);
			}
			break;
		case OutputStream:
			break;
		}
		
		if (faucet instanceof IPipelineItem) {
			((IPipelineItem) faucet).consume(this);
		}
		
		try {
			inputStream = (InputStream) faucet.getSource(ContentType.ByteStream);
			int count;
			byte[] buffer = new byte[8 * 1024];
			if (logger.isTraceEnabled())
				logger.trace("Using the reader " + inputStream + " in the mux sink");
			while ((count = inputStream.read(buffer)) != -1) {
				if (logger.isTraceEnabled()) {
					String readInput = ByteArrayUtil.getHexDump(buffer, 0, count);
					logger.trace("Read " + readInput + " from input stream " +
							inputStream);
				}
				outputStream.write(buffer, 0, count);
			}
		} catch (IOException ioe) {
			throw new XformationException("Unable to transform stream", ioe);
		}
	}

	@Override
	public Future<?> consumeAsynchronously() {
		return ExecutorsHelper.getInstance().submitInProc(new Runnable() {
			@Override
			public void run() {
				try {
					consume();
				} catch (Throwable t) {
					logger.error("Unable to consume faucet", t);
				}
			}
		});
	}

	@Override
	public void readFrom(DataInputStream in) throws IOException,
			IllegalAccessException, InstantiationException {
		this.id = in.readUTF();
		this.faucet = null;
		this.hookedUp = false;
		this.type = Type.File;
		this.outputStream = null;
		this.inputStream = null;
		this.path = in.readUTF();
	}

	@Override
	public void writeTo(DataOutputStream out) throws IOException {
		if (type != Type.File)
			throw new XformationException(
					"Only StreamSinks of type Type.File can be streamed.");
		
		out.writeUTF(id);
		out.writeUTF(path);
	}

}
