/*******************************************************************************
 * Copyright (c) 2009 Pafsanias Ftakas.
 *
 * This file is part of DistProc.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * DistProc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Eclipse Public License v1.0 for more details.
 *
 * Contributors:
 *     Pafsanias Ftakas - initial API and implementation
 *******************************************************************************/
package com.koutra.dist.proc.sink;

import java.io.BufferedWriter;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.io.Writer;
import java.util.concurrent.BlockingDeque;
import java.util.concurrent.Future;
import java.util.concurrent.LinkedBlockingDeque;

import org.apache.log4j.Logger;

import com.koutra.dist.proc.model.ContentType;
import com.koutra.dist.proc.model.IFaucet;
import com.koutra.dist.proc.model.IMuxSink;
import com.koutra.dist.proc.model.IPipelineItem;
import com.koutra.dist.proc.model.ISinkTemplate;
import com.koutra.dist.proc.model.XformationException;
import com.koutra.dist.proc.util.ExecutorsHelper;

public class WriterMuxSink implements IMuxSink {
	private static final Logger logger = Logger.getLogger(WriterMuxSink.class);
	
	protected enum Type {
		File,
		Writer;
	};
	
	protected static class DequePayload {
		public IFaucet faucet;
		public Reader reader;
		
		@Override
		public boolean equals(Object obj) {
			if (!(obj instanceof DequePayload))
				return false;
			
			DequePayload other = (DequePayload) obj;
			if (reader == null && other.reader != null)
				return false;
			return faucet.equals(other.faucet) && reader.equals(other.reader);
		}
	}
	
	protected String id;
	protected IFaucet faucetTemplate;
	protected BlockingDeque<DequePayload> deque;
	protected IFaucet faucet;
	protected boolean hookedUp;
	protected Type type;
	protected Writer writer;
	protected String path;
	protected String outputCharset;
	
	/**
	 * @deprecated Use any of the initializing constructors instead.
	 */
	public WriterMuxSink() { }
	
	public WriterMuxSink(String id, Writer writer) {
		this.id = id;
		this.faucetTemplate = null;
		this.deque = new LinkedBlockingDeque<DequePayload>();
		this.hookedUp = false;
		this.type = Type.Writer;
		this.writer = writer;
		this.path = null;
		this.outputCharset = null;
	}

	public WriterMuxSink(String id, String path, String outputCharset) {
		this.id = id;
		this.faucetTemplate = null;
		this.deque = new LinkedBlockingDeque<DequePayload>();
		this.hookedUp = false;
		this.type = Type.File;
		this.writer = null;
		this.path = path;
		this.outputCharset = outputCharset;
	}

	@Override
	public String getId() {
		return id;
	}

	@Override
	public IFaucet getFaucet() {
		DequePayload payload = deque.peek();
		return payload == null ? null : payload.faucet;
	}

	@Override
	public boolean supportsInput(ContentType contentType) {
		switch (contentType) {
		case CharStream:
			return true;
		case ByteStream:
		case XML:
		case ResultSet:
		default:
			return false;
		}
	}

	@Override
	public void hookupFaucet(IFaucet faucet) {
		if (!faucet.supportsOutput(ContentType.CharStream))
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' must support the CharStream content type.");
		if (!(faucet instanceof ISinkTemplate))
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' must implement the ISinkTemplate interface.");
		if (!((ISinkTemplate) faucet).isClone()) {
			if (faucetTemplate != null && faucetTemplate == faucet) {
				// Effectively a NO-OP
				return;
			}
			if (faucetTemplate != null)
				throw new XformationException("Hooking a faucet template in an " +
						"already set mux faucet");
			faucetTemplate = faucet;
			faucetTemplate.hookupSink(this);
			return;
		}
		if (hookedUp && faucet == this.faucet) {
			// Effectively a NO-OP.
			return;
		}
		
		this.faucet = faucet;
		this.hookedUp = true;
		
		faucet.hookupSink(this);
	}
	
	@Override
	public void registerSource(Object source) {
		Reader reader = (Reader) source;
		
		DequePayload payload = new DequePayload();
		payload.faucet = faucet;
		payload.reader = reader;

		if (deque.contains(payload))
			return;
		
		if (logger.isTraceEnabled())
			logger.trace("Registering reader: " + reader + " with the mux deque");
		
		deque.addLast(payload);
		
		if (reader != null) {
			if (faucet instanceof IPipelineItem) {
				((IPipelineItem) faucet).consume(this);
			}
		}
	}

	@Override
	public String dumpPipeline() {
		DequePayload payload = deque.peekFirst();
		return getClass().getName() + ": " +
			(payload == null ? null : payload.reader) + "->" + writer;
	}

	@Override
	public void dispose() {
		switch (type) {
		case File:
			try {
				writer.close();
			} catch (IOException e) {
				throw new XformationException("Unable to close writer", e);
			}
			break;
		case Writer:
			// Caller owns the writer. It is their responsibility to close it.
			break;
		}
	}

	@Override
	public void consume() {
		if (!hookedUp && faucetTemplate == null)
			throw new XformationException("Sink has not been set up correctly: " +
					"faucet has not been set");
		
		switch (type) {
		case File:
			try {
				if (outputCharset == null)
					writer = new BufferedWriter(new FileWriter(path));
				else
					writer = new BufferedWriter(new OutputStreamWriter(
							new FileOutputStream(path), outputCharset));
			} catch (IOException e) {
				throw new XformationException("Unable to create output writer", e);
			}
			break;
		case Writer:
			break;
		}
		
		if (!hookedUp) {
			// This is an initial run up the chain. Propagate the consumption of the
			// event. This will start filling up the faucet and reader deque.
			if (faucetTemplate instanceof IPipelineItem) {
				((IPipelineItem) faucetTemplate).consume(this);
			}
		}
			
		try {
			int count;
			char[] buffer = new char[8 * 1024];
			while (true) {
				DequePayload payload = null;
				try {
					payload = deque.takeFirst();
				} catch (InterruptedException ie) { }
				if (payload == null)
					break;
				
				IFaucet faucet = payload.faucet;
				Reader reader = payload.reader;
				
				if (logger.isTraceEnabled())
					logger.trace("Removed reader: " + reader + " from the deque.");
				
				// If a null has been sent down to the deque then we must exit.
				if (reader == null)
					break;
			
				if (logger.isTraceEnabled())
					logger.trace("Using the reader " + reader + " in the mux sink");
				while ((count = reader.read(buffer)) != -1) {
					if (logger.isTraceEnabled())
						logger.trace("Read '" + new String(buffer, 0, count) +
								"' from reader " + reader + " in the mux sink");
					writer.write(buffer, 0, count);
				}
				
				faucet.dispose();
			}
			
			writer.close();
			faucetTemplate.dispose();
		} catch (IOException ioe) {
			throw new XformationException("Unable to transform char stream", ioe);
		}
	}

	@Override
	public Future<?> consumeAsynchronously() {
		return ExecutorsHelper.getInstance().submitInProc(new Runnable() {
			@Override
			public void run() {
				try {
					consume();
				} catch (Throwable t) {
					logger.error("Unable to consume faucet", t);
				}
			}
		});
	}

	@Override
	public void readFrom(DataInputStream in) throws IOException,
			IllegalAccessException, InstantiationException {
		this.id = in.readUTF();
		this.faucetTemplate = null;
		this.deque = new LinkedBlockingDeque<DequePayload>();
		this.faucet = null;
		this.hookedUp = false;
		this.type = Type.File;
		this.writer = null;
		this.path = in.readUTF();
		this.outputCharset = in.readUTF();
	}

	@Override
	public void writeTo(DataOutputStream out) throws IOException {
		if (type != Type.File)
			throw new XformationException(
					"Only WriterMuxSinks of type Type.File can be streamed.");
		
		out.writeUTF(id);
		out.writeUTF(path);
		out.writeUTF(outputCharset);
	}

}
